# coding: utf-8
"""
Implementation for:
- InvalidCredentialsError: Exception raised when the credentials are invalid
"""


class InvalidCredentialsError(Exception):
    """
    Exception raised when the credentials are invalid
    """
    pass
