# coding=utf-8
"""
UEVaultManager setup file.
"""
import datetime

year = datetime.date.today().year

__name__ = 'UEVaultManager'
__version__ = '1.6.4'
# 0 Pegasus Seiya
# 1 Dragon Shiryu
# 2 Cygnus Hyoga
# 3 Andromeda Shun
# 4 Phoenix Ikki
# 5 Leo Aiolia
# 5 Virgo Shaka
__codename__ = 'Virgo+4'
# 6 Libra Dohko
# 7 Scorpio Milo
# 8 Sagittarius Aiolos
# 9 Capricorn Shura
# 10 Aquarius Camus
# 11 Pisces Aphrodite
__author__ = 'Laurent Ongaro'
__author_email__ = 'laurent@gameamea.com'
__description__ = 'Free and open-source replacement for the Epic Games Launcher application, mainly to manage the assets for Unreal Engine'
__copyright__ = f'{year} {__author__}'
__license__ = 'GPL-3'
__url__ = 'https://github.com/LaurentOngaro/UEVaultManager'
